# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch
from nemo.collections.common.data.fallback import FallbackDataset


class DummyDataset(torch.utils.data.Dataset):
    def __getitem__(self, item):
        return item


def test_fallback_dataset():
    dset = DummyDataset()
    assert dset[0] == 0
    assert dset[None] is None

    fallback = FallbackDataset(dset)
    assert fallback[0] == 0
    assert fallback[None] == 0
    assert fallback[1] == 1
    assert fallback[None] == 1

    fallback = FallbackDataset(dset)
    assert fallback[0] == 0
    assert fallback[None] == 0
    assert fallback[None] == 0

    fallback = FallbackDataset(dset)
    assert fallback[None] is None
    assert fallback[0] == 0
    assert fallback[None] == 0
