# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# fmt: off
from nemo.collections.asr.parts.submodules.adapters.attention_adapter_mixin import AttentionAdapterModuleMixin
from nemo.collections.asr.parts.submodules.adapters.multi_head_attention_adapter_module import (
    MHAResidualAddAdapterStrategy,
    MHAResidualAddAdapterStrategyConfig,
    MultiHeadAttentionAdapter,
    MultiHeadAttentionAdapterConfig,
    PositionalEncodingAdapter,
    PositionalEncodingAdapterConfig,
    RelPositionalEncodingAdapter,
    RelPositionalEncodingAdapterConfig,
    RelPositionMultiHeadAttentionAdapter,
    RelPositionMultiHeadAttentionAdapterConfig,
)
from nemo.collections.asr.parts.submodules.adapters.transformer_multi_head_attention_adapter_module import (
    TransformerMultiHeadAttentionAdapter,
    TransformerMultiHeadAttentionAdapterConfig,
)

# fmt: on
