# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Any

from cloudpickle import dump, load

from nemo.lightning.io.artifact.base import Artifact


class PickleArtifact(Artifact[Any]):
    def dump(self, instance, absolute_dir: Path, relative_dir: Path) -> Path:
        relative_file = self.file_path(relative_dir)
        with open(Path(absolute_dir) / relative_file, "wb") as f:
            dump(value, f)

        return relative_file

    def load(self, path: Path) -> Any:
        with open(self.file_path(path), "rb") as f:
            return load(f)

    def file_path(self, path: Path) -> Path:
        return path / self.attr + ".pkl"
